@@ -1,5 +1,6 @@
 /* Copyright (C) 2002-2005 RealVNC Ltd.  All Rights Reserved.
  * Copyright 2004-2005 Cendio AB.
+ * Copyright 2017 Peter Astrand <astrand@cendio.se> for Cendio AB
  * 
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,24 +24,14 @@
 #include <ctype.h>
 #include <string.h>
 
+#include <os/Mutex.h>
+
 #include <rfb/util.h>
 #include <rfb/Configuration.h>
 #include <rfb/LogWriter.h>
 #include <rfb/Exception.h>
-#include <rfb/Threading.h>
 
-#ifdef __RFB_THREADING_IMPL
-// On platforms that support Threading, we use Locks to make getData safe
-#define LOCK_CONFIG Lock l(*configLock())
-rfb::Mutex* configLock_ = 0;
-static rfb::Mutex* configLock() {
-  if (!configLock_)
-    configLock_ = new rfb::Mutex;
-  return configLock_;
-}
-#else
-#define LOCK_CONFIG
-#endif
+#define LOCK_CONFIG os::AutoMutex a(mutex)
 
 #include <rdr/HexOutStream.h>
 #include <rdr/HexInStream.h>
@@ -88,7 +79,6 @@ bool Configuration::set(const char* name, int len,
         strncasecmp(current->getName(), name, len) == 0)
     {
       bool b = current->setParam(val);
-      current->setHasBeenSet(); 
       if (b && immutable) 
 	current->setImmutable();
       return b;
@@ -113,7 +103,6 @@ bool Configuration::set(const char* config, bool immutable) {
     while (current) {
       if (strcasecmp(current->getName(), config) == 0) {
         bool b = current->setParam();
-	current->setHasBeenSet(); 
         if (b && immutable) 
 	  current->setImmutable();
         return b;
@@ -178,11 +167,28 @@ void Configuration::list(int width, int nameWidth) {
 }
 
 
+bool Configuration::remove(const char* param) {
+  VoidParameter *current = head;
+  VoidParameter **prevnext = &head;
+
+  while (current) {
+    if (strcasecmp(current->getName(), param) == 0) {
+      *prevnext = current->_next;
+      return true;
+    }
+    prevnext = &current->_next;
+    current = current->_next;
+  }
+
+  return false;
+}
+
+
 // -=- VoidParameter
 
 VoidParameter::VoidParameter(const char* name_, const char* desc_,
 			     ConfigurationObject co)
-  : immutable(false), _hasBeenSet(false), name(name_), description(desc_)
+  : immutable(false), name(name_), description(desc_)
 {
   Configuration *conf = NULL;
 
@@ -197,9 +203,12 @@ VoidParameter::VoidParameter(const char* name_, const char* desc_,
 
   _next = conf->head;
   conf->head = this;
+
+  mutex = new os::Mutex();
 }
 
 VoidParameter::~VoidParameter() {
+  delete mutex;
 }
 
 const char*
@@ -226,16 +235,6 @@ VoidParameter::setImmutable() {
   immutable = true;
 }
 
-void
-VoidParameter::setHasBeenSet() {
-  _hasBeenSet = true;
-} 
-
-bool
-VoidParameter::hasBeenSet() {
-  return _hasBeenSet;
-}
-
 // -=- AliasParameter
 
 AliasParameter::AliasParameter(const char* name_, const char* desc_,
@@ -339,7 +338,7 @@ bool
 IntParameter::setParam(const char* v) {
   if (immutable) return true;
   vlog.debug("set %s(Int) to %s", getName(), v);
-  int i = atoi(v);
+  int i = strtol(v, NULL, 0);
   if (i < minValue || i > maxValue)
     return false;
   value = i;
@@ -380,7 +379,7 @@ StringParameter::StringParameter(const char* name_, const char* desc_,
   : VoidParameter(name_, desc_, co), value(strDup(v)), def_value(v)
 {
   if (!v) {
-    fprintf(stderr,"Default value <null> for %s not allowed\n",name_);
+    vlog.error("Default value <null> for %s not allowed",name_);
     throw rfb::Exception("Default value <null> not allowed");
   }
 }
